from ...schema import Schema

from .text import TextPrompt
from .password import PasswordPrompt
from .confirm import ConfirmPrompt
from .choice import ChoicePrompt
from .checkbox import CheckboxPrompt
from .number import NumberPrompt
from .date import DatePrompt
from .editor import EditorPrompt
from .key import KeyPrompt


class Prompt:
    """A class for prompting the user for input."""

    @staticmethod
    def text(
        message: str,
        schema: Schema | None = None,
        id: str | None = None,
    ) -> TextPrompt:
        """Creates a text prompt where the user can input a text string."""
        return TextPrompt(message, schema, id)

    @staticmethod
    def password(
        message: str,
        schema: Schema | None = None,
        id: str | None = None,
    ) -> PasswordPrompt:
        """Creates a password prompt where the user can input a text string masked with '*'."""
        return PasswordPrompt(message, schema, id)

    @staticmethod
    def confirm(
        message: str,
        schema: Schema | None = None,
        id: str | None = None,
    ) -> ConfirmPrompt:
        """Creates a confirm prompt where the user can confirm an action, either yes or no."""
        return ConfirmPrompt(message, schema, id)

    @staticmethod
    def choice(
        message: str,
        schema: Schema | None = None,
        id: str | None = None,
    ) -> ChoicePrompt:
        """Creates a choice prompt where the user can select from a list of choices."""
        return ChoicePrompt(message, schema, id)

    @staticmethod
    def checkbox(
        message: str,
        schema: Schema | None = None,
        id: str | None = None,
    ) -> CheckboxPrompt:
        """Creates a checkbox prompt where the user can select multiple choices from a list of choices."""
        return CheckboxPrompt(message, schema, id)

    @staticmethod
    def number(
        message: str,
        schema: Schema | None = None,
        id: str | None = None,
    ) -> NumberPrompt:
        """Creates a number prompt where the user can input a number."""
        return NumberPrompt(message, schema, id)

    @staticmethod
    def date(
        message: str,
        schema: Schema | None = None,
        id: str | None = None,
    ) -> DatePrompt:
        """Creates a date prompt where the user can input a date."""
        return DatePrompt(message, schema, id)

    @staticmethod
    def editor(
        message: str,
        schema: Schema | None = None,
        id: str | None = None,
    ) -> EditorPrompt:
        """Creates an editor prompt where the user can input multiple lines of text, along with syntax highlighting if specified."""
        return EditorPrompt(message, schema, id)

    @staticmethod
    def keypress(
        message: str,
        schema: Schema | None = None,
        id: str | None = None,
    ) -> KeyPrompt:
        """Creates a keypress prompt where the user can input multiple keys."""
        return KeyPrompt(message, schema, id)



import re
from colorama import Fore, Back
import sys
import signal
from .constants import Keys


def wrap(text: str, width: int) -> list:
    """Wraps text to the desired width.

    Args:
        text (str): Text to wrap.
        width (int): Width to wrap text to.

    Returns:
        list: A list of lines as they appear in the block of wrapped text.
    """
    result = []

    lines = text.split("\n")

    for line in lines:
        if line.strip() == "":
            # Preserve empty lines (including lines with only whitespace)
            result.append("")
            continue

        current_line = ""
        current_line_visible_length = 0
        words = re.findall(r"\S+\s*", line)

        for word in words:
            word_visible_length = len(strip_ansi(word))

            if current_line_visible_length + word_visible_length > width:
                if current_line:
                    result.append(current_line.rstrip())
                current_line = ""
                current_line_visible_length = 0

            current_line += word
            current_line_visible_length += word_visible_length

        if current_line:
            result.append(current_line.rstrip())

    return result


def strip_ansi(text: str) -> str:
    """
    Remove ANSI escape sequences from a string.

    This function removes all ANSI escape sequences, including color codes,
    cursor movements, and other control sequences.

    Args:
        text (str): The input string containing ANSI escape sequences.

    Returns:
        str: The input string with all ANSI escape sequences removed.
    """
    ansi_escape = re.compile(
        r"""\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])""",
        re.VERBOSE,
    )
    return ansi_escape.sub("", text)


def rgb_to_ansi(r: int = 255, g: int = 255, b: int = 255, fg: bool = True) -> str:
    color_code = 16 + 36 * int(r / 255 * 5) + 6 * int(g / 255 * 5) + int(b / 255 * 5)
    return f'\x1b[{"38" if fg else "48"};5;{color_code}m'


def colorize(string: str, color: dict[str, tuple | str]) -> str:
    """Applied certain ANSI escape codes that apply color.

    Args:
        string (str): String to color.
        color (dict[str, tuple  |  str]): A dictionary that contains keys 'foreground' and 'background' with values for colors in either RGB as a tuple or a color string recognized by the colorama library.

    Returns:
        str: The inputted string with color formatting.
    """

    fg_color = color.get("foreground")
    bg_color = color.get("background")

    fg_code = (
        rgb_to_ansi(*fg_color, fg=True)
        if isinstance(fg_color, tuple) and len(fg_color) == 3
        else (getattr(Fore, fg_color.upper(), "") if isinstance(fg_color, str) else "")
    )
    bg_code = (
        rgb_to_ansi(*bg_color, fg=False)
        if isinstance(bg_color, tuple) and len(bg_color) == 3
        else (getattr(Back, bg_color.upper(), "") if isinstance(bg_color, str) else "")
    )

    return f"{fg_code}{bg_code}{string}\x1b[0m"


def strip_unsafe_objs(
    object: any, repr_id: str
) -> str | int | bool | float | list | dict | tuple:
    """Sanitizes objects for unsafe objects.

    Args:
        object (any): Object to sanitize.
        repr_id (str): An ID that identifies canonically altered strings.

    Returns:
        str | int | bool | float | list | dict | tuple: A sanitized object.
    """
    if isinstance(object, (str, int, float, bool)):
        return object
    elif isinstance(object, list):
        return [strip_unsafe_objs(item, repr_id) for item in object]
    elif isinstance(object, dict):
        return {key: strip_unsafe_objs(value, repr_id) for key, value in object.items()}
    elif isinstance(object, tuple):
        return tuple(strip_unsafe_objs(item, repr_id) for item in object)
    else:
        return f"{repr_id}{repr(object)}"


def strip_repr_id(string: str, repr_id: str) -> str:
    """Removes repr IDs from strings.

    Args:
        string (str): String to operate on.
        repr_id (str): Repr ID to remove.

    Returns:
        str: String without repr IDs.
    """
    return re.sub(
        rf'"{repr_id}(.*?)"', r"\1", re.sub(rf"'{repr_id}(.*?)'", r"\1", string)
    )


def if_space(string: str, space: int = 0) -> str:
    """If space allows, return the given string.

    Args:
        string (str): String in question.
        space (int, optional): Space permitted. Defaults to 0.

    Returns:
        str: The original string if space permits, otherwise "".
    """
    return string if space >= len(string) else ""


def get_key() -> str:
    def handle_interrupt(signum, frame):
        raise KeyboardInterrupt()

    if sys.platform.startswith("win"):
        import msvcrt

        # Set up the interrupt handler
        signal.signal(signal.SIGINT, handle_interrupt)

        try:
            while True:
                if msvcrt.kbhit():
                    char = msvcrt.getch().decode("utf-8")
                    if char == Keys.CTRLC:  # Ctrl+C
                        raise KeyboardInterrupt()
                    return char
        finally:
            # Reset the interrupt handler
            signal.signal(signal.SIGINT, signal.SIG_DFL)

    else:
        import termios
        import tty

        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(fd)
            # Set up the interrupt handler
            signal.signal(signal.SIGINT, handle_interrupt)

            while True:
                char = sys.stdin.read(1)
                if char == Keys.CTRLC:  # Ctrl+C
                    raise KeyboardInterrupt()
                if char == Keys.ESCAPE:
                    # Handle escape sequences (e.g., arrow keys)
                    next_char = sys.stdin.read(1)
                    if next_char == "[":
                        last_char = sys.stdin.read(1)
                        return f"\x1b[{last_char}"
                return char
        finally:
            # Reset terminal settings and interrupt handler
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
            signal.signal(signal.SIGINT, signal.SIG_DFL)



import os
import sys
import json
import base64
from typing import List, Optional, Dict, Any, Union, Callable


# Environment variable that indicates we're in shell control mode
SHELL_CONTROL_ENV = "ZENIF_SHELL_CONTROL"

# Shell types
SHELL_BASH = "bash"
SHELL_ZSH = "zsh"
SHELL_FISH = "fish"
SHELL_UNKNOWN = "unknown"


def is_shell_control_mode() -> bool:
    """Check if the CLI is running in shell control mode."""
    return os.environ.get(SHELL_CONTROL_ENV, "0") == "1"


def detect_shell() -> str:
    """Detect the type of shell the parent process is using."""
    shell_path = os.environ.get("SHELL", "")
    
    if "bash" in shell_path:
        return SHELL_BASH
    elif "zsh" in shell_path:
        return SHELL_ZSH
    elif "fish" in shell_path:
        return SHELL_FISH
    else:
        return SHELL_UNKNOWN


def format_shell_command(cmd: str, shell_type: Optional[str] = None) -> str:
    """
    Format a shell command for evaluation by the parent shell.
    
    Args:
        cmd: The shell command to format
        shell_type: Override detected shell type
        
    Returns:
        A formatted command string ready for shell evaluation
    """
    if shell_type is None:
        shell_type = detect_shell()
    
    # Add special marker so the wrapper can identify shell commands
    return f"ZENIF_SHELL_CMD:{cmd}"


def cd(directory: str) -> str:
    """
    Generate a shell command to change the current directory.
    
    Args:
        directory: The directory to change to
        
    Returns:
        A formatted shell command string
    """
    # Ensure the directory path is absolute
    if not os.path.isabs(directory):
        directory = os.path.abspath(directory)
    
    return format_shell_command(f"cd {directory}")


def set_env(name: str, value: str) -> str:
    """
    Generate a shell command to set an environment variable.
    
    Args:
        name: The name of the environment variable
        value: The value to set
        
    Returns:
        A formatted shell command string
    """
    # Different shells have different export syntax
    shell_type = detect_shell()
    
    if shell_type == SHELL_FISH:
        return format_shell_command(f"set -x {name} {value}")
    else:
        return format_shell_command(f"export {name}=\"{value}\"")


def unset_env(name: str) -> str:
    """
    Generate a shell command to unset an environment variable.
    
    Args:
        name: The name of the environment variable to unset
        
    Returns:
        A formatted shell command string
    """
    shell_type = detect_shell()
    
    if shell_type == SHELL_FISH:
        return format_shell_command(f"set -e {name}")
    else:
        return format_shell_command(f"unset {name}")


def run_raw(command: str) -> str:
    """
    Generate a shell command to execute arbitrary shell code.
    Use with caution as this could be a security risk if not properly sanitized.
    
    Args:
        command: The raw shell command to execute
        
    Returns:
        A formatted shell command string
    """
    return format_shell_command(command)


def shell_output_wrapper(result: Any) -> None:
    """
    Output function that handles shell control commands vs normal output.
    
    Args:
        result: The result to output. If it's a shell control command,
               it will be formatted for shell evaluation.
    """
    if isinstance(result, str) and result.startswith("ZENIF_SHELL_CMD:"):
        # Extract the actual command
        cmd = result[len("ZENIF_SHELL_CMD:"):]
        print(f"__ZENIF_SHELL_EVAL__{cmd}__ZENIF_SHELL_EVAL_END__")
    else:
        # Regular output
        print(result)


# Shell wrapper function definition for users to include in their shell config
def get_shell_wrapper_script(shell_type: Optional[str] = None) -> str:
    """
    Get the shell wrapper script that should be sourced by the user.
    
    Args:
        shell_type: Force a specific shell type, or auto-detect if None
        
    Returns:
        A shell script that defines the shell wrapper function
    """
    if shell_type is None:
        shell_type = detect_shell()
    
    if shell_type == SHELL_FISH:
        return """
function zenif
    set -l output (env ZENIF_SHELL_CONTROL=1 command zenif $argv)
    for line in $output
        if string match -q -r '^__ZENIF_SHELL_EVAL__.*__ZENIF_SHELL_EVAL_END__$' -- $line
            set -l cmd (string replace -r '^__ZENIF_SHELL_EVAL__(.*?)__ZENIF_SHELL_EVAL_END__$' '$1' -- $line)
            eval $cmd
        else
            echo $line
        end
    end
end
"""
    else:  # Default for bash/zsh and others
        return """
zenif() {
    local output=$(ZENIF_SHELL_CONTROL=1 command zenif "$@")
    local line
    
    while IFS= read -r line; do
        if [[ $line =~ ^__ZENIF_SHELL_EVAL__.*__ZENIF_SHELL_EVAL_END__$ ]]; then
            # Extract and evaluate the shell command
            local cmd="${line#__ZENIF_SHELL_EVAL__}"
            cmd="${cmd%__ZENIF_SHELL_EVAL_END__}"
            eval "$cmd"
        else
            echo "$line"
        fi
    done <<< "$output"
}
"""


from .base import BasePrompt
from ...schema import Schema, ListF
from ...constants import Keys, Cursor

from colorama import Fore, Style


class CheckboxPrompt(BasePrompt):
    def __init__(
        self,
        message: str,
        schema: Schema | None = None,
        id: str | None = None,
    ):

        super().__init__(message, schema, id)
        self._choices = []

        # Check if the field is a ListF
        if schema and not isinstance(self.field, ListF):
            field_type = type(self.field).__name__
            error_message = (
                f"CheckboxPrompt requires a ListF field, but got {field_type}"
            )
            raise TypeError(error_message)

    def choices(self, *choices: str | list[str]) -> "CheckboxPrompt":
        """Set the choices for the prompt."""
        flattened = []
        for choice in choices:
            if isinstance(choice, list):
                flattened.extend(str(item) for item in choice)
            else:
                flattened.append(str(choice))
        self._choices = flattened
        return self

    def ask(self) -> list[str]:
        """Prompt the user for input."""
        selected = [False] * len(self._choices)
        current = 0

        controls = "↑/↓ to navigate, Space to select, Enter to confirm"

        check = "•"  # "X"
        hover = "○"  # check
        hovercheck = "●"  # f"\x1b[4m{check}\x1b[0m"

        print()
        print()

        i = True
        while True:
            for j, (choice, is_selected) in enumerate(zip(self._choices, selected)):
                print()
                if j == current:
                    print(f"{Fore.YELLOW}{Style.DIM}{hover}{Style.NORMAL}", end="")
                else:
                    print(f"{Fore.YELLOW} ", end="")
                print(
                    f"\r{f"{Fore.YELLOW}{hovercheck if j == current else check}{Style.RESET_ALL}" if is_selected else Cursor.right(1)} {Fore.YELLOW}{Style.DIM}{choice}{Fore.RESET}",
                    end="",
                )

            key = ""
            if i:
                i = False
                result = [
                    choice
                    for choice, is_selected in zip(self._choices, selected)
                    if is_selected
                ]
                error = self.validate(result)

                print(Cursor.up(len(self._choices) + 1), end="")
                self._print_prompt(self.message, error=f"{error if error else ""}\n")
                print(
                    f"\r{Fore.RESET}{Style.DIM}  {controls}{Cursor.down(len(self._choices) - 1)}"
                )
            else:
                key = self._get_key()
            if key == " ":
                selected[current] = not selected[current]

            result = [
                choice
                for choice, is_selected in zip(self._choices, selected)
                if is_selected
            ]
            error = self.validate(result)

            print(Cursor.up(len(self._choices) + 1), end="")
            self._print_prompt(self.message, error=f"{error if error else ""}\n")
            print(
                f"\r{Fore.RESET}{Style.DIM}  {controls}{Cursor.down(len(self._choices) - 1)}"
            )

            if key == Keys.ENTER and not error:
                for _ in range(len(self._choices) + 1):
                    print(Cursor.lclear() + Cursor.up(1) + Cursor.lclear(), end="")
                self._print_prompt(
                    self.message,
                    (
                        ", ".join(map(str, result[:-1])) + f", and {result[-1]}"
                        if len(result) > 1
                        else str(result[0])
                    ),
                )
                print()
                return result
            elif key == Keys.UP and current > 0:
                current -= 1
            elif key == Keys.DOWN and current < len(self._choices) - 1:
                current += 1

            print(Cursor.up(len(self._choices) + 1))  # Move cursor up to redraw choices



from typing import Any
from ..utils import strip_ansi, wrap
from .template import TemplateEngine

from datetime import datetime, UTC
from shutil import get_terminal_size as tsize
from threading import current_thread
from io import StringIO
from re import sub
from copy import deepcopy
from colorama import Style

class Ruleset:

    def __init__(self, rules: dict[str, Any], defaults: dict[str, Any]):
        self._rules = self._merge_with_defaults(deepcopy(rules), deepcopy(defaults))
        self.dict = self.Dict(self._rules)
        for category, settings in self._rules.items():
            setattr(
                self,
                category,
                type(
                    f"{category.title().replace('_', '')}Rules", (), deepcopy(settings)
                )(),
            )

    def _merge_with_defaults(self, rules: dict[str, Any], defaults: dict[str, Any]):
        merged = defaults.copy()
        for category, settings in rules.items():
            if category in merged:
                merged[category].update(settings)
            else:
                merged[category] = settings
        return merged

    class Dict:
        def __init__(self, rules):
            self.all = rules
            for category, settings in rules.items():
                setattr(self, category, deepcopy(settings))


class BaseHandler:

    def __init__(self, defaults: dict[str, Any]):
        self.output_streams = []
        self.stream_rulesets = {}
        self.template_engine = TemplateEngine()
        self.previous_timestamp = None
        self.defaults = deepcopy(defaults)

    def add(self, stream: str | object, ruleset: dict[str, Any] = None):
        if stream not in self.output_streams:
            self.output_streams.append(stream)
            if ruleset:
                self.stream_rulesets[stream] = Ruleset(deepcopy(ruleset), self.defaults)
            else:
                self.stream_rulesets[stream] = Ruleset({}, self.defaults)
        return stream

    def remove(self, stream: str | object):
        if stream in self.output_streams:
            self.output_streams.remove(stream)
            if stream in self.stream_rulesets:
                del self.stream_rulesets[stream]

    def modify(
        self, stream: str | object, ruleset: dict[str, Any], use_original: bool = False
    ):
        if stream not in self.stream_rulesets:
            raise ValueError(f"Stream {stream} not found in handler.")

        if use_original:
            new_rules = deepcopy(self.defaults)
            new_rules.update(deepcopy(ruleset))
            self.stream_rulesets[stream] = Ruleset(new_rules, self.defaults)
        else:
            current_rules = deepcopy(self.stream_rulesets[stream]._rules)
            current_rules.update(deepcopy(ruleset))
            self.stream_rulesets[stream] = Ruleset(current_rules, self.defaults)

    def process_message(
        self,
        message: str,
        level: dict[str, str | int],
        metadata: dict[str, Any],
        ruleset: dict[str, Any],
    ):
        terminal_width = tsize().columns

        context = {
            "timestamp": self._define_timestamp(ruleset),
            "filename": metadata.get("file_name", ""),
            "wrapfunc": metadata.get("wrapping_func", ""),
            "linenum": metadata.get("line_number", ""),
            "level": level["name"],
        }

        log_line = self.template_engine.process(
            ruleset.log_line.format, context, level["name"]
        )
        log_line = sub(r"\x1b\[(\d+)C", lambda m: " " * int(m.group(1)), log_line)
        log_line = log_line if ruleset.formatting.ansi else strip_ansi(log_line)

        message_space = terminal_width - self.template_engine.processed.length
        message_indent = self.template_engine.processed.length

        log_metadata = self._generate_metadata(
            metadata, ruleset, message_space, terminal_width
        )

        formatted_message = message if ruleset.formatting.ansi else strip_ansi(message)

        log_line += log_metadata

        lines = wrap(formatted_message, message_space)

        log_output = StringIO()

        log_output.write(
            f"{log_line}{''.join([f'{line}\n' if i == 0 else f'\x1b[{message_indent}C{line}{Style.RESET_ALL}\n' for i, line in enumerate(lines)])}"
        )

        return log_output.getvalue()

    def _define_timestamp(self, ruleset: dict[str, Any]):
        now = datetime.now(UTC if ruleset.timestamps.use_utc else None)
        timestamp = now.strftime("%H:%M:%S")
        if timestamp == self.previous_timestamp and not ruleset.timestamps.always_show:
            return f"\x1b[{len(timestamp)}C"
        self.previous_timestamp = timestamp
        return timestamp

    def _generate_metadata(
        self,
        metadata: dict[str, Any],
        ruleset: dict[str, Any],
        message_space: int,
        terminal_width: int,
    ):
        if not ruleset.metadata.show_metadata:
            return ""

        metadata_items = []
        if ruleset.metadata.include_timestamp:
            metadata_items.append(f"[tms: {self._define_timestamp(ruleset)}]")
        if ruleset.metadata.include_level_name:
            metadata_items.append(f"[lvl: {metadata['level']}]")
        if ruleset.metadata.include_thread_name:
            metadata_items.append(f"[thr: {current_thread().name}]")
        if ruleset.metadata.include_file_name:
            metadata_items.append(f"[fl: {metadata['file_name']}]")
        if ruleset.metadata.include_wrapping_function:
            metadata_items.append(f"[wfc: {metadata['wrapping_func']}]")
        if ruleset.metadata.include_function:
            metadata_items.append(f"[fnc: {metadata['function']}]")
        if ruleset.metadata.include_line_number:
            metadata_items.append(f"[ln: {metadata['line_number']}]")
        if ruleset.metadata.include_value_count:
            metadata_items.append(f"[vlc: {metadata['value_count']}]")

        metadata_str = " ".join(metadata_items)
        metadata_space = terminal_width - message_space - len(strip_ansi(metadata_str))
        return f"{Style.RESET_ALL}{Style.DIM}{metadata_str.rjust(metadata_space)}{Style.RESET_ALL}"


class StreamHandler(BaseHandler):

    def __init__(self, defaults: dict[str, Any]):
        super().__init__(defaults)
        self.output_streams = []
        self.stream_rulesets = {}

    def add(self, stream: object, ruleset: dict[str, Any] = None):
        if stream not in self.output_streams:
            self.output_streams.append(stream)
            if ruleset:
                self.stream_rulesets[stream] = Ruleset(deepcopy(ruleset), self.defaults)
            else:
                self.stream_rulesets[stream] = Ruleset({}, self.defaults)
        return stream

    def modify(
        self, stream: object, ruleset: dict[str, Any], use_original: bool = False
    ):
        if stream not in self.stream_rulesets:
            raise ValueError(f"Stream {stream} not found in handler.")

        if use_original:
            new_rules = deepcopy(self.defaults)
            new_rules.update(deepcopy(ruleset))
            self.stream_rulesets[stream] = Ruleset(new_rules, self.defaults)
        else:
            current_rules = deepcopy(self.stream_rulesets[stream].dict.all)
            current_rules.update(deepcopy(ruleset))
            self.stream_rulesets[stream] = Ruleset(current_rules, self.defaults)

    def write(
        self, message: str, level_dict: dict[str, str | int], metadata: dict[str, Any]
    ):
        for stream in self.output_streams:
            ruleset = self.stream_rulesets.get(stream)
            if ruleset:
                if not self._should_log(message, level_dict["level"], ruleset):
                    continue
                processed_message = self.process_message(
                    message, level_dict, metadata, ruleset
                )
                stream.write(processed_message)
            else:
                stream.write(message)
            stream.flush()

    def _should_log(self, message: str, level: int, ruleset: dict[str, Any]):
        if level < ruleset.filtering.min_level:
            return False
        if any(
            substring in message for substring in ruleset.filtering.exclude_messages
        ):
            return False
        if ruleset.filtering.include_only_messages and not any(
            substring in message
            for substring in ruleset.filtering.include_only_messages
        ):
            return False
        return True


class FileHandler(BaseHandler):
    def __init__(self, defaults: dict[str, Any]):
        super().__init__(defaults)
        self.file_streams = {}
        self.file_rulesets = {}

    def add(
        self, file_path: str, ruleset: dict[str, Any] | None = None, reset: bool = False
    ):
        if file_path not in self.file_streams or reset:
            mode = "w" if reset else "a"
            self.file_streams[file_path] = open(file_path, mode)
            if ruleset:
                self.file_rulesets[file_path] = Ruleset(
                    deepcopy(ruleset), self.defaults
                )
            else:
                self.file_rulesets[file_path] = Ruleset({}, self.defaults)
        return file_path

    def remove(self, file_path: str):
        if file_path in self.file_streams:
            self.file_streams[file_path].close()
            del self.file_streams[file_path]
            if file_path in self.file_rulesets:
                del self.file_rulesets[file_path]

    def modify(
        self, file_path: str, ruleset: dict[str, Any], use_original: bool = False
    ):
        if file_path in self.file_rulesets:
            if use_original:
                new_rules = deepcopy(self.defaults)
                new_rules.update(deepcopy(ruleset))
                self.file_rulesets[file_path] = Ruleset(new_rules, self.defaults)
            else:
                current_rules = deepcopy(self.file_rulesets[file_path]._rules)
                current_rules.update(deepcopy(ruleset))
                self.file_rulesets[file_path] = Ruleset(current_rules, self.defaults)

    def write(
        self, message: str, level_dict: dict[str, str | int], metadata: dict[str, Any]
    ):
        for file_path, file_stream in self.file_streams.items():
            ruleset = self.file_rulesets.get(file_path)
            if ruleset:
                if not self._should_log(message, level_dict["level"], ruleset):
                    continue
                processed_message = self.process_message(
                    message, level_dict, metadata, ruleset
                )
                file_stream.write(strip_ansi(processed_message))
            else:
                file_stream.write(strip_ansi(message))
            file_stream.flush()

    def _should_log(self, message: str, level: int, ruleset: dict[str, Any]):
        if level < ruleset.filtering.min_level:
            return False
        if any(
            substring in message for substring in ruleset.filtering.exclude_messages
        ):
            return False
        if ruleset.filtering.include_only_messages and not any(
            substring in message
            for substring in ruleset.filtering.include_only_messages
        ):
            return False
        return True

    def reset(self, file_path: str):
        if file_path in self.file_streams:
            self.remove(file_path)
            self.add(file_path, reset=True)


class Streams:
    def __init__(self, defaults: dict[str, Any]):
        self.file = FileHandler(defaults)
        self.normal = StreamHandler(defaults)


class FHGroup:
    def __init__(self, file_handler: FileHandler, *items):
        self.file_handler = file_handler
        self.file_paths = []
        self.add(*items)

    def unwrap_fhgroups(self, items):
        new = []
        for item in items:
            if isinstance(item, str):
                new.append(item)
            elif isinstance(item, FHGroup):
                new.extend(self.unwrap_fhgroups(item.file_paths))
        return new

    def add(self, *items):
        items = self.unwrap_fhgroups(items)
        for item in items:
            if item not in self.file_paths:
                self.file_paths.append(item)
                self.file_handler.add(item)
        return [*items]

    def remove(self, *items: str):
        items = self.unwrap_fhgroups(items)
        for file_path in items:
            if file_path in self.file_paths:
                self.file_paths.remove(file_path)
                self.file_handler.remove(file_path)
        return [*items]

    def reset(self):
        for file_path in self.file_paths:
            self.file_handler.remove(file_path)
            self.file_handler.add(file_path, reset=True)

    def modify(self, ruleset: dict[str, Any], use_original: bool = False):
        for file_path in self.file_paths:
            self.file_handler.modify(file_path, ruleset, use_original)

    def remove_all(self):
        for file_path in self.file_paths:
            self.file_handler.remove(file_path)
        self.file_paths.clear()


class SHGroup:

    def __init__(self, stream_handler: StreamHandler, *items):
        self.stream_handler = stream_handler
        self.streams = []
        self.add(*items)

    def unwrap_shgroups(self, items):
        new = []
        for item in items:
            if isinstance(item, SHGroup):
                new.extend(item.streams)
            else:
                new.append(item)
        return new

    def add(self, *items):
        items = self.unwrap_shgroups(items)
        for item in items:
            if item not in self.streams:
                self.streams.append(item)
                self.stream_handler.add(item)
        return [*items]

    def remove(self, *items: str):
        items = self.unwrap_shgroups(items)
        for item in items:
            if item in self.streams:
                self.streams.remove(item)
                self.stream_handler.remove(item)
        return [*items]

    def modify(self, ruleset: dict[str, Any], use_original: bool = False):
        for stream in self.streams:
            self.stream_handler.modify(stream, ruleset, use_original)

    def remove_all(self):
        for stream in self.streams:
            self.stream_handler.remove(stream)
        self.streams.clear()



from .base import BasePrompt
from ...schema import Schema, StringF
from ...constants import Keys, Cursor

from colorama import Fore, Style


class ChoicePrompt(BasePrompt):
    def __init__(
        self,
        message: str,
        schema: Schema | None = None,
        id: str | None = None,
    ):
        super().__init__(message, schema, id)
        self._choices = []

        # Check if the field is a StringF
        if schema and not isinstance(self.field, StringF):
            field_type = type(self.field).__name__
            error_message = (
                f"ChoicePrompt requires a StringF field, but got {field_type}"
            )
            raise TypeError(error_message)

    def choices(self, *choices: str | list[str]) -> "ChoicePrompt":
        """Set the choices for the prompt."""
        flattened = []
        for choice in choices:
            if isinstance(choice, list):
                flattened.extend(str(item) for item in choice)
            else:
                flattened.append(str(choice))
        self._choices = flattened
        return self

    def ask(self) -> str:
        """Prompt the user for input."""
        current = 0

        controls = "↑/↓ to navigate, Enter to confirm"

        check = "→"

        print(
            f"{Fore.GREEN}? {Fore.CYAN}{self.message}:{Fore.RESET}\n{Style.DIM}  {controls}"
        )
        while True:
            for i, choice in enumerate(self._choices):
                if i > 0:
                    print()
                if i == current:
                    print(
                        f"{Fore.YELLOW}{Style.NORMAL}{check} {choice}{Fore.RESET}",
                        end="",
                    )
                else:
                    print(f"{Fore.YELLOW}{Style.DIM}  {choice}{Fore.RESET}", end="")

            key = self._get_key()
            if key == Keys.ENTER:  # Enter key
                result = self._choices[current]
                error = self.validate(result or "")
                if not error:
                    for _ in range(len(self._choices) + 1):
                        print(Cursor.lclear() + Cursor.up(1) + Cursor.lclear(), end="")
                    self._print_prompt(self.message, result)
                    print()  # Move to next line
                    return result
                else:
                    for _ in range(len(self._choices) + 1):
                        print(Cursor.lclear() + Cursor.up(1) + Cursor.lclear(), end="")
                    self._print_prompt(self.message, error=error)
                    print()
                    print(
                        f"{Fore.GREEN}? {Fore.CYAN}{self.message}:{Fore.RESET}\n{Style.DIM}  {controls}"
                    )
            elif key == Keys.UP and current > 0:  # Up arrow
                current -= 1
            elif key == Keys.DOWN and current < len(self._choices) - 1:  # Down arrow
                current += 1

            print(Cursor.up(len(self._choices)))  # Move cursor up to redraw choices



from textwrap import dedent, indent
from shutil import get_terminal_size as tsize
from typing import Any
from colorama import Fore, Back, Style
from .parameters import Parameter


class HelpFormatter:
    @staticmethod
    def format_cli_help(cli_name: str, commands: dict[str, Any]) -> str:
        """
        Format help text for the entire CLI application in a visually pleasing style.
        Commands with aliases will display like "fetch, f".
        """
        lines = []
        lines.append(
            f"{Back.BLUE}{Fore.BLACK}  {cli_name} <command> [args]  {Style.RESET_ALL}"
        )
        lines.append("")

        lines.append(
            f"{Back.BLUE}{Fore.BLACK}  {'Command':<20} {'Description'.ljust(tsize().columns - 23)}{Style.RESET_ALL}"
        )

        # Format each primary command. If the command has aliases (stored on _aliases), list them.
        for name, command in sorted(commands.items()):
            # Get the aliases that were registered on the command.
            aliases = getattr(command, "_aliases", [])
            if aliases:
                # Build a display string like "fetch, f, another_alias"
                display_name = f"{name}, {', '.join(aliases)}"
            else:
                display_name = name

            doc = (
                command.__doc__.strip()
                if command.__doc__
                else "No description available."
            )
            first_line = doc.split("\n")[0]
            lines.append(
                f"{Fore.BLUE}  {display_name:<20} {first_line}{Style.RESET_ALL}"
            )
        return "\n".join(lines)

    @staticmethod
    def format_command_help(command_name: str, command: Any) -> str:
        """
        Format help text for a single command.
        If the command has aliases, list them next to the command name.
        """
        lines = []
        # Retrieve aliases registered on the command (if any)
        aliases = getattr(command, "_aliases", [])
        if aliases:
            header = (
                f"{Back.BLUE}{Fore.BLACK} {', '.join([command_name] + aliases)}"
                f"{Fore.BLUE}:{Style.RESET_ALL}{Fore.BLUE}  "
                f"{dedent(command.__doc__ if command.__doc__ else 'No description').strip().split('\n')[0]}"
                f"{Style.RESET_ALL}"
            )
        else:
            header = (
                f"{Back.BLUE}{Fore.BLACK} {command_name}{Fore.BLUE}:{Style.RESET_ALL}{Fore.BLUE}  "
                f"{dedent(command.__doc__ if command.__doc__ else 'No description').strip().split('\n')[0]}"
                f"{Style.RESET_ALL}"
            )
        lines.append(header)

        # Safely get the command documentation or a default message
        doc = command.__doc__ or "No description"
        doc_lines = dedent(doc).strip().split("\n")[1:]
        for line in doc_lines:
            lines.append(f"{Fore.BLUE}{Style.DIM}{line}{Style.RESET_ALL}")

        # Now, format the parameters (and their aliases, if set on the parameter objects).
        cli_params = getattr(command, "_cli_params", {})
        cli_aliases = getattr(command, "_cli_aliases", {})
        for param, alias in cli_aliases.items():
            if param in cli_params:
                if not alias.startswith("-"):
                    alias = f"-{alias}" if len(alias) == 1 else f"--{alias}"
                cli_params[param].alias = alias

        # Add a default help flag for the command.
        cli_params["--help"] = Parameter(
            param_name="help",
            kind="flag",
            help="Show this help menu",
            default=False,
            alias="h",
        )

        if cli_params:
            for param in sorted(cli_params.values(), key=lambda p: p.param_name):
                name = param.cli_name
                if name.startswith("--"):
                    name = "" + name
                elif name.startswith("-"):
                    name = " " + name
                else:
                    name = "  " + name
                if param.alias:
                    name += f" ({param.alias})"
                kind = param.kind.capitalize()
                default = param.default if param.default is not None else ""
                description = param.help
                lines.append("")
                lines.append(
                    f"{Fore.BLUE}{name}  {Style.DIM}{kind}\n{Fore.BLUE}{Style.DIM}  {description}{f"\n{Fore.RESET}  Defaults to {default}" if len(str(default)) > 0 and param.kind == "option" else ""}{Style.RESET_ALL}"
                )
        else:
            lines.append(
                f"{Fore.BLUE}No arguments defined for this command.{Style.RESET_ALL}"
            )

        return "\n".join(lines)
